@@ -12,7 +12,11 @@ from django.utils.html import escape
 from jsonpatch import JsonPatch, multidict
 
 from pistoke.nakyma import WebsocketNakyma
-from pistoke.tyokalut import csrf_tarkistus, json_viestiliikenne
+from pistoke.protokolla import WebsocketAliprotokolla
+from pistoke.tyokalut import (
+  CsrfKattely,
+  JsonLiikenne,
+)
 
 
 class WebsocketYhteys(WebsocketNakyma):
@@ -67,15 +71,15 @@ class WebsocketYhteys(WebsocketNakyma):
     if type(self) is __class__:
       raise TypeError(f'{__class__} on abstrakti!')
     super().__init__(*args, **kwargs)
-    #self.toimintojono = set()
+    self.toimintojono = set()
     # def __init__
 
+  websocket_protokolla = 'django-synkroni'
+
   def websocket_protokolla_json(self):
     ''' Palauta JSON-muotoinen tuettujen protokollien luettelo. '''
-    if not getattr(self.websocket, 'protokolla', None):
-      return 'null'
     return escape(self.json_koodain().encode(
-      self.websocket.protokolla
+      [self.websocket_protokolla]
     ))
     # def websocket_protokolla_json
 
@@ -93,18 +97,17 @@ class WebsocketYhteys(WebsocketNakyma):
   async def kasittele_toiminto(self, request, *, toiminto_id, **kwargs):
     try:
       vastaus = await self.suorita_toiminto(**kwargs)
-    # pylint: disable=broad-except
-    except BaseException as exc:
-      # Kuitataan toiminto suoritetuksi poikkeuksesta huolimatta.
-      await request.send({
-        'toiminto_id': toiminto_id,
-      })
-      traceback.print_exception(exc)
-    else:
       await request.send({
         'toiminto_id': toiminto_id,
         **(vastaus or {})
       })
+    # pylint: disable=broad-except
+    except BaseException:
+      # Kuitataan toiminto suoritetuksi poikkeuksesta huolimatta.
+      await request.send({
+        'toiminto_id': toiminto_id,
+      })
+      raise
     # async def kasittele_toiminto
 
   async def kasittele_saapuva_sanoma(self, request, sanoma):
@@ -112,11 +115,16 @@ class WebsocketYhteys(WebsocketNakyma):
     if 'toiminto_id' in sanoma:
       # Komento: suorita taustalla.
       # Useat peräkkäin saapuvat toiminnot ajetaan samanaikaisesti.
-      await self.kasittele_toiminto(request, **sanoma)
-      #toiminto = asyncio.ensure_future(
-      #  self.kasittele_toiminto(request, **sanoma)
-      #)
-      #self.toimintojono.add(toiminto)
+      toiminto = asyncio.ensure_future(
+        self.kasittele_toiminto(request, **sanoma)
+      )
+      self.toimintojono.add(toiminto)
+      @toiminto.add_done_callback
+      def done(task):
+        try:
+          task.result()
+        except BaseException:
+          traceback.print_exc()
 
     else:
       # Json-paikkaus: toteuta tässä.
@@ -144,38 +152,31 @@ class WebsocketYhteys(WebsocketNakyma):
         await self.kasittele_saapuva_sanoma(request, sanoma)
         # while True
     finally:
-      pass
-      #for kesken in self.toimintojono:
-      #  kesken.cancel()
-      #tulokset = await asyncio.gather(
-      #  *self.toimintojono,
-      #  return_exceptions=True
-      #)
-      #for tulos in tulokset:
-      #  if isinstance(tulos, BaseException):
-      #    raise tulos
+      for kesken in self.toimintojono:
+        kesken.cancel()
+      await asyncio.gather(*self.toimintojono)
       # finally
     # async def _websocket
 
   async def websocket(self, request, *args, **kwargs):
     ''' JSON-ohjaimet valitaan pyyntökohtaisesti. '''
     # pylint: disable=no-self-argument, arguments-differ
-    @json_viestiliikenne(
+    @WebsocketAliprotokolla(self.websocket_protokolla)
+    @JsonLiikenne(
       # Käytetään luokkakohtaisesti määriteltyä JSON-protokollaa
       # viestinnässä selaimen kanssa.
       loads={'cls': self.json_latain},
       dumps={'cls': self.json_koodain},
     )
-    @csrf_tarkistus(
+    @CsrfKattely(
       csrf_avain='csrfmiddlewaretoken',
       virhe_avain='virhe'
     )
-    async def websocket(self, request, *args, **kwargs):
+    async def websocket(request, *args, **kwargs):
       # pylint: disable=protected-access
       return await self._websocket(request, *args, **kwargs)
       # async def websocket
-    return await websocket(self, request, *args, **kwargs)
+    return await websocket(request, *args, **kwargs)
     # def websocket
-  websocket.protokolla = 'django-synkroni'
 
   # class WebsocketYhteys
